/*
 * The contents of this file are subject to the AOLserver Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://aolserver.lcs.mit.edu/.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is AOLserver Code and related documentation
 * distributed by AOL.
 * 
 * The Initial Developer of the Original Code is America Online,
 * Inc. Portions created by AOL are Copyright (C) 1999 America Online,
 * Inc. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public License (the "GPL"), in which case the
 * provisions of GPL are applicable instead of those above.  If you wish
 * to allow use of your version of this file only under the terms of the
 * GPL and not to allow others to use your version of this file under the
 * License, indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by the GPL.
 * If you do not delete the provisions above, a recipient may use your
 * version of this file under either the License or the GPL.
 */


/*
 * tclpages.c --
 *
 *      Run a TCL page possibly from a file cache.
 */

static char rcsid[] = "@(#) $Id: tclpages.c,v 1.5 1999/12/15 05:55:30 mayoff Exp $";

#include "ns.h"
#include <sys/stat.h>
#include <unistd.h>

int Ns_ModuleVersion = 1;

typedef struct Key {
    dev_t dev;
    ino_t ino;
} Key;

typedef struct TclPage {
    int          isCached;
    time_t       mtime;
    size_t       size;
    int          refcnt;
    char         bytes[1]; /* actually bytes[st_size+1] */
} TclPage;

/*
 * Local functions defined in this file
 */

static int TclPageGet(void *ignored, Ns_Conn *conn);

/*
 * Static variables defined in this file
 */

static Ns_ModLogHandle  tclPageLog;
static Ns_Cache        *tclPageCache;
static int              tclPageCacheMaxEntry;

static void RegisterMap(char *hServer, char *map);
static void RegisterMaps(char *hServer, char *moduleConfigPath);
static void MaybeCreateCache(char *moduleConfigPath);
static TclPage *GetPage(char *filename);
static TclPage *GetPageFromCache(char *filename, struct stat *st);
static TclPage *GetPageFromDisk(char *filename, struct stat *st, int isCached);
static int RunTcl(Ns_Conn *conn, char *url, char *code);
static void FreeTclPage(TclPage *page);

/* 
 *==========================================================================
 * Initialization functions
 *==========================================================================
 */


/*
 *----------------------------------------------------------------------
 * Ns_ModuleInit --
 *
 *      Initialization function called at server start up.
 *
 * Results:
 *      NS_OK.
 *
 * Side effects:
 *      File cache may be created.
 *
 *----------------------------------------------------------------------
 */

int
Ns_ModuleInit(char *hServer, char *hModule)
{
    char   *serverConfigPath;
    char   *moduleConfigPath;
    int     enableTclPages;
    int     rc;

    Ns_ModLogRegister("tclPages", &tclPageLog);

    serverConfigPath = Ns_ConfigGetPath(hServer, NULL, NULL);
    moduleConfigPath = Ns_ConfigGetPath(hServer, hModule, NULL);

    rc = Ns_ConfigGetBool(serverConfigPath,
	"EnableTclPages", &enableTclPages);

    if (rc == NS_FALSE || enableTclPages == NS_FALSE) {
	Ns_ModLog(Notice, tclPageLog, "Tcl pages NOT enabled");
    } else {

	Ns_ModLog(Notice, tclPageLog, "Tcl pages ENABLED");

	MaybeCreateCache(moduleConfigPath);

	RegisterMaps(hServer, moduleConfigPath);
    }

    return NS_OK;
}

/*--------------------------------------------------------------------*/

static void
RegisterMap(char *hServer, char *map)
{
    Ns_ModLog(Notice, tclPageLog, "registering map \"%s\"", map);
    Ns_RegisterRequest(hServer, "GET", map, TclPageGet, NULL, NULL, 0);
    Ns_RegisterRequest(hServer, "HEAD", map, TclPageGet, NULL, NULL, 0);
    Ns_RegisterRequest(hServer, "POST", map, TclPageGet, NULL, NULL, 0);
}

/*--------------------------------------------------------------------*/

static void
RegisterMaps(char *hServer, char *moduleConfigPath)
{
    Ns_Set *set;
    int     mapCount;
    int     i, setSize;
    char   *map;

    set = Ns_ConfigGetSection(moduleConfigPath);
    mapCount = 0;
    if (set != NULL) {
	setSize = Ns_SetSize(set);
	for (i = 0; i < setSize; i++) {
	    if (STRIEQ(Ns_SetKey(set, i), "map")) {
		RegisterMap(hServer, Ns_SetValue(set, i));
		mapCount++;
	    }
	}
    }

    if (mapCount == 0) {
	RegisterMap(hServer, "*.tcl");
    }
}

/*--------------------------------------------------------------------*/

static void
MaybeCreateCache(char *moduleConfigPath)
{
    int rc;
    int enableCache;
    int maxCacheSize;

    rc = Ns_ConfigGetBool(moduleConfigPath, "cache", &enableCache);
    if (rc == NS_FALSE)
	enableCache = NS_FALSE;

    rc = Ns_ConfigGetInt(moduleConfigPath, "cachemaxsize",
	&maxCacheSize);
    if (rc == NS_FALSE)
	maxCacheSize = 5 * 1024 * 1024;

    rc = Ns_ConfigGetInt(moduleConfigPath, "cachemaxentry",
	&tclPageCacheMaxEntry);
    if (rc == NS_FALSE)
	tclPageCacheMaxEntry = 32 * 1024;

    if (enableCache == NS_TRUE && maxCacheSize < tclPageCacheMaxEntry) {
	Ns_ModLog(Warning, tclPageLog,
	    "cache max size (%d) less than cache entry max size (%d);"
	    " disabling cache", maxCacheSize, tclPageCacheMaxEntry);
	enableCache = NS_FALSE;
    }

    if (enableCache == NS_FALSE) {

	Ns_ModLog(Notice, tclPageLog, "Tcl page caching NOT enabled");

	tclPageCacheMaxEntry = 0;
	tclPageCache = NULL;

    } else {

	Ns_ModLog(Notice, tclPageLog, "Tcl page caching ENABLED"
	    " (size = %d, max cached page size = %d)",
	    maxCacheSize, tclPageCacheMaxEntry);

	tclPageCache = Ns_CacheCreateSz("tclpages",
	    sizeof(Key) / sizeof(int), maxCacheSize,
	    (Ns_Callback *)FreeTclPage);
    }
}


/*
 *==========================================================================
 * Page retrieval functions
 *==========================================================================
 */


/*
 *----------------------------------------------------------------------
 * TclPageGet --
 *
 *      Run a TCL file.
 *
 * Results:
 *      Return NS_OK for success or NS_ERROR for failure.
 *
 * Side effects:
 *      Contents of file may be cached in file cache.
 *
 *----------------------------------------------------------------------
 */

static int
TclPageGet(void *ignored, Ns_Conn *conn)
{
    Ns_DString   ds;
    int          retval;
    TclPage     *page = NULL;

    Ns_DStringInit(&ds);

    if (Ns_UrlToFile(&ds, nsServer, conn->request->url) == NS_OK) {
	page = GetPage(Ns_DStringValue(&ds));
    }

    if (page != NULL) {
	retval = RunTcl(conn, conn->request->url, page->bytes);

	if (page->isCached) {
	    Ns_CacheLock(tclPageCache);
	    FreeTclPage(page);
	    Ns_CacheUnlock(tclPageCache);
	} else {
	    FreeTclPage(page);
	}
    }

    else {
	retval = Ns_ConnReturnNotFound(conn);
    }

    Ns_DStringFree(&ds);
    return retval;
}


/*
 *----------------------------------------------------------------------
 * GetFile --
 *
 *      Get the contents of the named file, possibly from the file
 *      cache.
 *
 * Results:
 *      Return a pointer to the file contents.  Return by reference
 *      a pointer to the function which releases the file contents.
 *      Log an error and return NULL if there was a problem.
 *
 * Side effects:
 *      Contents of file may be cached in file cache.
 *
 *----------------------------------------------------------------------
 */

static TclPage *
GetPage(char *filename)
{
    struct stat  st;
    TclPage     *page = NULL;

    if (stat(filename, &st) < 0) {
	Ns_ModLog(Debug, tclPageLog, "stat(%s) failed: %s",
		  filename, strerror(errno));
    }

    else if (!S_ISREG(st.st_mode)) {
	Ns_ModLog(Error, tclPageLog, "%s is not a regular file",
		  filename);
    }

    else if (st.st_size > (size_t)tclPageCacheMaxEntry) {
	page = GetPageFromDisk(filename, &st, 0);
    }

    else {
	page = GetPageFromCache(filename, &st);
    }

    return page;
}

/*--------------------------------------------------------------------*/

static TclPage *
GetPageFromCache(char *filename, struct stat *st)
{
    Key       key;
    Ns_Entry *entry;
    int       isNew;
    TclPage  *page;

    Ns_ModLog(Debug, tclPageLog, "looking for \"%s\" in cache",
	filename);

    key.dev = st->st_dev;
    key.ino = st->st_ino;

    Ns_CacheLock(tclPageCache);

    entry = Ns_CacheCreateEntry(tclPageCache, (char *)&key, &isNew);
    page = (TclPage *)Ns_CacheGetValue(entry);

    if (page != NULL) {
	Ns_ModLog(Debug, tclPageLog, "found cache entry");
	if (page->mtime != st->st_mtime
	    || page->size != st->st_size)
	{
	    Ns_ModLog(Debug, tclPageLog, "cache entry is stale");
	    page = NULL;
	}
    }

    if (page == NULL) {
	page = GetPageFromDisk(filename, st, 1);
	if (page != NULL) {
	    Ns_CacheSetValueSz(entry, (void *)page, page->size);
	}
    }

    page->refcnt++;
    Ns_CacheUnlock(tclPageCache);

    return page;
}

/*--------------------------------------------------------------------*/

static TclPage *
GetPageFromDisk(char *filename, struct stat *st, int isCached)
{
    TclPage     *page;
    int          fd, rc;

    Ns_ModLog(Debug, tclPageLog, "loading \"%s\" from disk%s",
	filename, isCached ? " for cache" : "");

    fd = open(filename, O_RDONLY);
    if (fd < 0) {
	Ns_ModLog(Error, tclPageLog, "open(%s) failed: %s",
		  filename, strerror(errno));
	page = NULL;
    }

    else {
	if (isCached) {
	    page = Ns_CacheMalloc(tclPageCache,
				 sizeof *page + st->st_size);
	} else {
	    page = Ns_Malloc(sizeof *page + st->st_size);
	}
	page->isCached = isCached;
	page->size = st->st_size;
	page->mtime = st->st_mtime;
	page->refcnt = 1;
	page->bytes[st->st_size] = 0;
	rc = read(fd, page->bytes, st->st_size);
	if (rc != st->st_size) {
	    if (rc < 0) {
		Ns_ModLog(Error, tclPageLog, "read(%s) failed: %s",
			  filename, strerror(errno));
	    }
	    else {
		Ns_ModLog(Error, tclPageLog,
		    "read(%s): short read - wanted %d, got %d",
		    filename, st->st_size, rc);
	    }
	    FreeTclPage(page);
	    page = NULL;
	}
	close(fd);
    }

    return page;
}


/*
 *----------------------------------------------------------------------
 * RunTcl --
 *
 *      Run some Tcl code.
 *
 * Results:
 *      NS_OK or NS_ERROR.
 *
 * Side effects:
 *      Depends on the Tcl code.
 *
 *----------------------------------------------------------------------
 */

static int
RunTcl(Ns_Conn *conn, char *url, char *code)
{
    int          retval = NS_ERROR;
    Tcl_Interp  *interp = Ns_GetConnInterp(conn);

    if (interp != NULL) {
	if (Tcl_Eval(interp, code) == TCL_OK) {
	    retval = NS_OK;
	} else {
	    char message[200];
	    sprintf(message, "\n    (url \"%.150s\" line %d)",
	      url, interp->errorLine);
	    Tcl_AddErrorInfo(interp, message);
	    Ns_TclLogError(interp);
	    retval = Ns_ConnReturnInternalError(conn);
	}
    }

    return retval;
}


/*--------------------------------------------------------------------*/

void
FreeTclPage(TclPage *page)
{
    if (page->isCached) {
	page->refcnt--;
	if (page->refcnt <= 0) {
	    Ns_CacheFree(tclPageCache, page);
	}
    } else {
	Ns_Free(page);
    }
}

